confattr.configfile module — confattr v0.7.0 documentation

# confattr.configfile module

This module defines the ConfigFile class which can be used to load and save config files.

class confattr.configfile.ArgumentParser(prog=None, usage=None, description=None, epilog=None, parents=[], formatter_class=<class 'argparse.HelpFormatter'>, prefix_chars='-', fromfile_prefix_chars=None, argument_default=None, conflict_handler='error', add_help=True, allow_abbrev=True, exit_on_error=True)

  Bases: `ArgumentParser`

  error(message: string)

    Prints a usage message incorporating the message to stderr and exits.

    If you override this in a subclass, it should not return – it should either exit or raise an exception.

class confattr.configfile.ConfigFile(*, notification_level: ~typing.Union[~confattr.config.Config[~confattr.configfile.NotificationLevel], ~confattr.configfile.NotificationLevel] = NotificationLevel.ERROR, appname: str, authorname: ~typing.Optional[str] = None, config_instances: dict[str, confattr.config.Config[typing.Any]] = {}, commands: ~typing.Optional[~collections.abc.Sequence[type[confattr.configfile.ConfigFileCommand]]] = None, formatter_class: type[argparse.HelpFormatter] = <class 'confattr.utils.HelpFormatter'>)

  Bases: `object`

  Read or write a config file.

  Parameters

    - notification_level – Messages of a lower priority are not passed to the callback registered with `set_ui_callback()`

    - appname – The name of the application, required for generating the path of the config file if you use `load()` or `save()` and as prefix of environment variable names

    - authorname – The name of the developer of the application, on MS Windows useful for generating the path of the config file if you use `load()` or `save()`

    - config_instances – The Config instances to load or save, defaults to `Config.instances`

    - commands – The `ConfigFileCommand`s allowed in this config file, if this is :const:`None`: use the return value of `ConfigFileCommand.get_command_types()`

    - formatter_class – Is used to clean up doc strings and wrap lines in the help

  COMMENT = '#'
  COMMENT_PREFIXES = ('"', '#')
  ENTER_GROUP_PREFIX = '['
  ENTER_GROUP_SUFFIX = ']'
  config_directory: str | None = None

    Override the config directory which is returned by `iter_user_site_config_paths()`. You should set either this attribute or `config_path` in your tests with `monkeypatch.setattr`. If the environment variable `APPNAME_CONFIG_DIRECTORY` is set this attribute is set to it’s value in the constructor (where `APPNAME` is the value which is passed as `appname` to the constructor but in all upper case letters and hyphens and spaces replaced by underscores.)

  config_id: Optional[ConfigId]
  config_name = 'config'

    The name of the config file used by `iter_config_paths()`. Can be changed with the environment variable `APPNAME_CONFIG_NAME` (where `APPNAME` is the value which is passed as `appname` to the constructor but in all upper case letters and hyphens and spaces replaced by underscores.).

  config_path: str | None = None

    Override the config file which is returned by `iter_config_paths()`. You should set either this attribute or `config_directory` in your tests with `monkeypatch.setattr`. If the environment variable `APPNAME_CONFIG_PATH` is set this attribute is set to it’s value in the constructor (where `APPNAME` is the value which is passed as `appname` to the constructor but in all upper case letters and hyphens and spaces replaced by underscores.)

  create_formatter() → HelpFormatterWrapper
  enter_group(ln: str) → bool

    Check if `ln` starts a new group and set `config_id` if it does.

    Parameters

      ln – The current line

    Returns

      `True` if `ln` starts a new group

  env_variables: list[str]
  get_app_dirs() → AppDirs

    Create or get a cached AppDirs instance with multipath support enabled.

    When creating a new instance, platformdirs, xdgappdirs and appdirs are tried, in that order. The first one installed is used. appdirs, the original of the two forks and the only one of the three with type stubs, is specified in pyproject.toml as a hard dependency so that at least one of the three should always be available. I am not very familiar with the differences but if a user finds that appdirs does not work for them they can choose to use an alternative with `pipx inject appname xdgappdirs|platformdirs`.

    These libraries should respect the environment variables `XDG_CONFIG_HOME` and `XDG_CONFIG_DIRS`.

  get_env_name(key: str) → str

    Convert the key of a setting to the name of the corresponding environment variable.

    Returns

      An all upper case version of `key` with all hyphens, dots and spaces replaced by underscores and `envprefix` prepended to the result. (`envprefix` is set in the constructor by first setting it to an empty str and then passing the value of `appname` to this method and appending an underscore.)

  get_help() → str

    A convenience wrapper around `write_help()` to return the help as a str instead of writing it to a file.

    This uses `HelpWriter`.

  get_help_config_id() → str
    Returns

      A help how to use `MultiConfig`. The return value still needs to be cleaned with `inspect.cleandoc()`.

  get_save_path() → str
    Returns

      The first existing and writable file returned by `iter_config_paths()` or the first path if none of the files are existing and writable.

  is_comment(ln: str) → bool

    Check if `ln` is a comment.

    Parameters

      ln – The current line

    Returns

      `True` if `ln` is a comment

  iter_config_paths() → Iterator[str]

    Iterate over all paths which are checked for config files, user specific first.

    Use this method if you want to tell the user where the application is looking for it’s config file. The first existing file yielded by this method is used by `load()`.

    The paths are generated by joining the directories yielded by `iter_user_site_config_paths()` with `ConfigFile.config_name`.

    If `config_path` has been set this method yields that path instead and no other paths.

  iter_user_site_config_paths() → Iterator[str]

    Iterate over all directories which are searched for config files, user specific first.

    The directories are based on `get_app_dirs()` unless `config_directory` has been set. If `config_directory` has been set it’s value is yielded and nothing else.

  load(*, env: bool = True) → None

    Load the first existing config file returned by `iter_config_paths()`.

    If there are several config files a user specific config file is preferred. If a user wants a system wide config file to be loaded, too, they can explicitly include it in their config file. :param _sphinx_paramlinks_confattr.configfile.ConfigFile.load.env: If true: call `load_env()` after loading the config file.

  load_env() → None

    Load settings from environment variables. The name of the environment variable belonging to a setting is generated with `get_env_name()`.

    Environment variables not matching a setting or having an invalid value are reported with `self.ui_notifier.show_error()`.

  load_file(fn: str) → None

    Load a config file and change the `Config` objects accordingly.

    Use `set_ui_callback()` to get error messages which appeared while loading the config file. You can call `set_ui_callback()` after this method without loosing any messages.

    Parameters

      fn – The file name of the config file (absolute or relative path)

  load_without_resetting_config_id(fn: str) → None
  parse_error(msg: str, **kw: Unpack[ParseSplittedLineKwargs]) → None

    Is called if something went wrong while trying to load a config file.

    This method is called when a `ParseException` or `MultipleParseExceptions` is caught. This method compiles the given information into an error message and calls `self.ui_notifier.show_error()`.

    Parameters

      - msg – The error message

      - line – The line where the error occured

      - line_number – The number of the line

      - file_name – The name of the config file from which the line has been read

  parse_line(line: str, **kw: Unpack[ParseLineKwargs]) → None
    Parameters

      - line – The line to be parsed

      - line_number – The number of the line, used in error messages

      - file_name – The name of the file from which ln was read (absolute or relative path), used in error messages and for relative imports

    `line_number` and `file_name` don’t need to be passed in case `ln` is not read from a config file but from a command line.

    `parse_error()` is called if something goes wrong, e.g. invalid key or invalid value.

  parse_splitted_line(ln_splitted: Sequence[str], **kw: Unpack[ParseSplittedLineKwargs]) → None
  quote(val: str) → str

    Quote a value if necessary so that it will be interpreted as one argument.

    The default implementation calls `readable_quote()`.

  save(**kw: Unpack[SaveKwargs]) → str

    Save the current values of all settings to the file returned by `get_save_path()`. Directories are created as necessary.

    Parameters

      - config_instances – Do not save all settings but only those given. If this is a `list` they are written in the given order. If this is a `set` they are sorted by their keys.

      - ignore – Do not write these settings to the file.

      - no_multi – Do not write several sections. For `MultiConfig` instances write the default values only.

      - comments – Write comments with allowed values and help.

    Returns

      The path to the file which has been written

  save_file(fn: str, **kw: Unpack[SaveKwargs]) → None

    Save the current values of all settings to a specific file.

    Parameters

      fn – The name of the file to write to. If this is not an absolute path it is relative to the current working directory.

    Raises

      FileNotFoundError – if the directory does not exist

    For an explanation of the other parameters see `save()`.

  save_to_open_file(f: TextIO, **kw: Unpack[SaveKwargs]) → None

    Save the current values of all settings to a file-like object by creating a `ConfigFileWriter` object and calling `save_to_writer()`.

    Parameters

      f – The file to write to

    For an explanation of the other parameters see `save()`.

  save_to_writer(writer: FormattedWriter, **kw: Unpack[SaveKwargs]) → None

    Save the current values of all settings.

    Ensure that all keyword arguments are passed with `set_save_default_arguments()`. Iterate over all `ConfigFileCommand` objects in `self.commands` and do for each of them:

    - set `should_write_heading` to `True` if `getattr(cmd.save, 'implemented', True)` is true for two or more of those commands or to `False` otherwise

    - call `save()`

  set_save_default_arguments(kw: SaveKwargs) → None

    Ensure that all arguments are given in `kw`.

  set_ui_callback(callback: Callable[[NotificationLevel, str | BaseException], None]) → None

    Register a callback to a user interface in order to show messages to the user like syntax errors or invalid values in the config file.

    Messages which occur before this method is called are stored and forwarded as soon as the callback is registered.

    Parameters

      ui_callback – A function to display messages to the user

  write_config_id(writer: FormattedWriter, config_id: ConfigId) → None

    Start a new group in the config file so that all following commands refer to the given `config_id`.

  write_help(writer: FormattedWriter) → None

class confattr.configfile.ConfigFileArgparseCommand(config_file: ConfigFile)

  Bases: `ConfigFileCommand`

  An abstract subclass of `ConfigFileCommand` which uses `argparse` to make parsing and providing help easier.

  You must implement the class method `init_parser()` to add the arguments to `parser`. Instead of `run()` you must implement `run_parsed()`. You don’t need to add a usage or the possible arguments to the doc string as `argparse` will do that for you. You should, however, still give a description what this command does in the doc string.

  You may specify `ConfigFileCommand.name`, `ConfigFileCommand.aliases` and `ConfigFileCommand.save()` like for `ConfigFileCommand`.

  get_help() → str

    Creates a help text which can be presented to the user by calling `parser.format_help()`. The return value of `ConfigFileCommand.write_help()` has been passed as `description` to the constructor of `ArgumentParser`, therefore `help`/the doc string are included as well.

  abstract init_parser(parser: ArgumentParser) → None
    Parameters

      parser – The parser to add arguments to. This is the same object like `parser`.

    This is an abstract method which must be implemented by subclasses. Use `ArgumentParser.add_argument()` to add arguments to `parser`.

  run(ln: Sequence[str], **kw: Unpack[ParseSplittedLineKwargs]) → None

    Process one line which has been read from a config file

    Raises

      - ParseException – if there is an error in the line (e.g. invalid syntax)

      - MultipleParseExceptions – if there are several errors in the same line

  abstract run_parsed(args: Namespace) → None

    This is an abstract method which must be implemented by subclasses.

class confattr.configfile.ConfigFileCommand(config_file: ConfigFile)

  Bases: `ABC`

  An abstract base class for commands which can be used in a config file.

  Subclasses must implement the `run()` method which is called when `ConfigFile` is loading a file. Subclasses should contain a doc string so that `get_help()` can provide a description to the user. Subclasses may set the `name` and `aliases` attributes to change the output of `get_name()` and `get_names()`.

  All subclasses are remembered and can be retrieved with `get_command_types()`. They are instantiated in the constructor of `ConfigFile`.

  add_help_to(formatter: HelpFormatterWrapper) → None

    Add the return value of `get_help_attr_or_doc_str()` to `formatter`.

  aliases: tuple[str, ...] | list[str]

    Alternative names which can be used in the config file.

  create_formatter() → HelpFormatterWrapper
  classmethod delete_command_type(cmd: type[confattr.configfile.ConfigFileCommand]) → None

    Delete `cmd` so that it is not returned anymore by `get_command_types()` and that it’s name can be used by another command. Do nothing if `cmd` has already been deleted.

  classmethod get_command_types() → tuple[type[confattr.configfile.ConfigFileCommand], ...]
    Returns

      All subclasses of `ConfigFileCommand` which have not been deleted with `delete_command_type()`

  get_help() → str
    Returns

      A help text which can be presented to the user.

    This is generated by creating a formatter with `create_formatter()`, adding the help to it with `add_help_to()` and stripping trailing new line characters from the result of `HelpFormatterWrapper.format_help()`.

    Most likely you don’t want to override this method but `add_help_to()` instead.

  get_help_attr_or_doc_str() → str
    Returns

      The `help` attribute or the doc string if `help` has not been set, cleaned up with `inspect.cleandoc()`.

  classmethod get_name() → str
    Returns

      The name which is used in config file to call this command.

    If `name` is set it is returned as it is. Otherwise a name is generated based on the class name.

  classmethod get_names() → Iterator[str]
    Returns

      Several alternative names which can be used in a config file to call this command.

    The first one is always the return value of `get_name()`. If `aliases` is set it’s items are yielded afterwards.

    If one of the returned items is the empty string this class is the default command and `run()` will be called if an undefined command is encountered.

  help: str

    A description which may be used by an in-app help. If this is not set `get_help()` uses the doc string instead.

  name: str

    The name which is used in the config file to call this command. Use an empty string to define a default command which is used if an undefined command is encountered. If this is not set `get_name()` returns the name of this class in lower case letters and underscores replaced by hyphens.

  abstract run(ln: Sequence[str], **kw: Unpack[ParseSplittedLineKwargs]) → None

    Process one line which has been read from a config file

    Raises

      - ParseException – if there is an error in the line (e.g. invalid syntax)

      - MultipleParseExceptions – if there are several errors in the same line

  save(writer: FormattedWriter, **kw: Unpack[SaveKwargs]) → None

    Write as many calls to this command as necessary to the config file in order to create the current state. There is the `config_file` attribute (which was passed to the constructor) which you can use to:

    - quote arguments with `ConfigFile.quote()`

    - get the comment character `ConfigFile.COMMENT`

    - call `ConfigFile.write_config_id`

    The default implementation does nothing.

  should_write_heading: bool = False

    If a config file contains only a single section it makes no sense to write a heading for it. This attribute is set by `ConfigFile.save_to_writer()` if there are several commands which implement the `save()` method. If you implement `save()` and this attribute is set then `save()` should write a section header. If `save()` writes several sections it should always write the headings regardless of this attribute.

class confattr.configfile.ConfigFileWriter(f: TextIO | None, prefix: str)

  Bases: `TextIOWriter`

  write_command(cmd: str) → None

    Write a config file command.

  write_heading(lvl: SectionLevel, heading: str) → None

    Write a heading.

    This object should not add an indentation depending on the section because if the indentation is increased the line width should be decreased in order to keep the line wrapping consistent. Wrapping lines is handled by `confattr.utils.HelpFormatter`, i.e. before the text is passed to this object. It would be possible to use `argparse.RawTextHelpFormatter` instead and handle line wrapping on a higher level but that would require to understand the help generated by argparse in order to know how far to indent a broken line. One of the trickiest parts would probably be to get the indentation of the usage right. Keep in mind that the term “usage” can differ depending on the language settings of the user.

    Parameters

      - lvl – How to format the heading

      - heading – The heading

  write_line(ln: str) → None

    Write a single line of documentation. `ln` may not contain a newline. If `ln` is empty it does not need to be prefixed with a comment character. Empty lines should be dropped if no other lines have been written before.

confattr.configfile.DEFAULT_COMMAND = ''

  If the name or an alias of `ConfigFileCommand` is this value that command is used by `ConfigFile.parse_splitted_line()` if an undefined command is encountered.

class confattr.configfile.FormattedWriter

  Bases: `ABC`

  abstract write_command(cmd: str) → None

    Write a config file command.

  abstract write_heading(lvl: SectionLevel, heading: str) → None

    Write a heading.

    This object should not add an indentation depending on the section because if the indentation is increased the line width should be decreased in order to keep the line wrapping consistent. Wrapping lines is handled by `confattr.utils.HelpFormatter`, i.e. before the text is passed to this object. It would be possible to use `argparse.RawTextHelpFormatter` instead and handle line wrapping on a higher level but that would require to understand the help generated by argparse in order to know how far to indent a broken line. One of the trickiest parts would probably be to get the indentation of the usage right. Keep in mind that the term “usage” can differ depending on the language settings of the user.

    Parameters

      - lvl – How to format the heading

      - heading – The heading

  abstract write_line(ln: str) → None

    Write a single line of documentation. `ln` may not contain a newline. If `ln` is empty it does not need to be prefixed with a comment character. Empty lines should be dropped if no other lines have been written before.

  write_lines(text: str) → None

    Write one or more lines of documentation.

class confattr.configfile.HelpWriter(f: TextIO | None)

  Bases: `TextIOWriter`

  write_command(cmd: str) → None

    Write a config file command.

  write_heading(lvl: SectionLevel, heading: str) → None

    Write a heading.

    This object should not add an indentation depending on the section because if the indentation is increased the line width should be decreased in order to keep the line wrapping consistent. Wrapping lines is handled by `confattr.utils.HelpFormatter`, i.e. before the text is passed to this object. It would be possible to use `argparse.RawTextHelpFormatter` instead and handle line wrapping on a higher level but that would require to understand the help generated by argparse in order to know how far to indent a broken line. One of the trickiest parts would probably be to get the indentation of the usage right. Keep in mind that the term “usage” can differ depending on the language settings of the user.

    Parameters

      - lvl – How to format the heading

      - heading – The heading

  write_line(ln: str) → None

    Write a single line of documentation. `ln` may not contain a newline. If `ln` is empty it does not need to be prefixed with a comment character. Empty lines should be dropped if no other lines have been written before.

class confattr.configfile.Include(config_file: ConfigFile)

  Bases: `ConfigFileArgparseCommand`

  Load another config file.

  This is useful if a config file is getting so big that you want to split it up or if you want to have different config files for different use cases which all include the same standard config file to avoid redundancy or if you want to bind several commands to one key which executes one command with ConfigFile.parse_line().

  By default the loaded config file starts with which ever config id is currently active. This is useful if you want to use the same values for several config ids: Write the set commands without a config id to a separate config file and include this file for every config id where these settings shall apply.

  After the include the config id is reset to the config id which was active at the beginning of the include because otherwise it might lead to confusion if the config id is changed in the included config file.

  init_parser(parser: ArgumentParser) → None
    Parameters

      parser – The parser to add arguments to. This is the same object like `parser`.

    This is an abstract method which must be implemented by subclasses. Use `ArgumentParser.add_argument()` to add arguments to `parser`.

  run(cmd: Sequence[str], **kw: Unpack[ParseSplittedLineKwargs]) → None

    Process one line which has been read from a config file

    Raises

      - ParseException – if there is an error in the line (e.g. invalid syntax)

      - MultipleParseExceptions – if there are several errors in the same line

  run_parsed(args: Namespace) → None

    This is an abstract method which must be implemented by subclasses.

exception confattr.configfile.MultipleParseExceptions(exceptions: Sequence[ParseException])

  Bases: `Exception`

  This is raised and caught inside of `ConfigFile` to communicate errors while parsing a config file where multiple settings are set in the same line. If you don’t intend to subclass `ConfigFile` you do not need to worry about this class.

class confattr.configfile.NotificationLevel(value)

  Bases: `SortedEnum`

  An enumeration.

  ERROR = 'error'
  INFO = 'info'
  descending: bool = False

exception confattr.configfile.ParseException

  Bases: `Exception`

  This is raised and caught inside of `ConfigFile` to communicate errors while parsing a config file. If you don’t intend to subclass `ConfigFile` you do not need to worry about this class.

class confattr.configfile.ParseLineKwargs

  Bases: `TypedDict`

  file_name: str
  line_number: int

class confattr.configfile.ParseSplittedLineKwargs

  Bases: `dict`

  file_name: str
  line: Required[str]
  line_number: int

class confattr.configfile.SaveKwargs

  Bases: `TypedDict`

  comments: bool
  config_instances: Iterable[Union[Config[Any], DictConfig[Any, Any]]]
  ignore: collections.abc.Iterable[Union[confattr.config.Config[Any], confattr.config.DictConfig[Any, Any]]] | None
  no_multi: bool

class confattr.configfile.SectionLevel(value)

  Bases: `SortedEnum`

  An enumeration.

  SECTION = 'section'

    Is used to separate different commands in `ConfigFile.write_help()` and `ConfigFileCommand.save()`

  SUB_SECTION = 'sub-section'

    Is used for subsections in `ConfigFileCommand.save()` such as the “data types” section in the help of the set command

  descending: bool = False

class confattr.configfile.Set(config_file: ConfigFile)

  Bases: `ConfigFileCommand`

  usage: set key1=val1 [key2=val2 …] \

    set key [=] val

  Change the value of a setting.

  In the first form set takes an arbitrary number of arguments, each argument sets one setting. This has the advantage that several settings can be changed at once. That is useful if you want to bind a set command to a key and process that command with ConfigFile.parse_line() if the key is pressed.

  In the second form set takes two arguments, the key and the value. Optionally a single equals character may be added in between as third argument. This has the advantage that key and value are separated by one or more spaces which can improve the readability of a config file.

  KEY_VAL_SEP = '='

    The separator which is used between a key and it’s value

  add_config_help(formatter: HelpFormatterWrapper, instance: Config[Any]) → None
  add_help_for_data_types(formatter: HelpFormatterWrapper, config_instances: Iterable[Config[object]]) → None
  add_help_to(formatter: HelpFormatterWrapper) → None

    Add the return value of `get_help_attr_or_doc_str()` to `formatter`.

  aliases: tuple[str, ...] | list[str]

    Alternative names which can be used in the config file.

  format_value(instance: Config[Any], config_id: Optional[ConfigId]) → str
    Parameters

      - instance – The config value to be saved

      - config_id – Which value to be written in case of a `MultiConfig`, should be `None` for a normal `Config` instance

    Returns

      A str representation to be written to the config file

    Convert the value of the `Config` instance into a str with `Config.format_value()`.

  get_data_type_name_to_help_map(config_instances: Iterable[Config[object]]) → dict[str, str]
    Parameters

      config_instances – All config values to be saved

    Returns

      A dictionary containing the type names as keys and the help as values

    The returned dictionary contains the help for all data types except enumerations which occur in `config_instances`. The help is gathered from the `help` attribute of the type or the str registered with `set_help_for_type()`.

  get_help_for_data_types(config_instances: Iterable[Config[object]]) → str
  help: str

    A description which may be used by an in-app help. If this is not set `get_help()` uses the doc string instead.

  help_for_types = {<class 'str'>: 'A text. If it contains spaces it must be wrapped in single or double quotes.', <class 'int'>: "\t\t\tAn integer number in python 3 syntax, as decimal (e.g. 42), hexadecimal (e.g. 0x2a), octal (e.g. 0o52) or binary (e.g. 0b101010).\n\t\t\tLeading zeroes are not permitted to avoid confusion with python 2's syntax for octal numbers.\n\t\t\tIt is permissible to group digits with underscores for better readability, e.g. 1_000_000.", <class 'float'>: 'A floating point number in python syntax, e.g. 23, 1.414, -1e3, 3.14_15_93.'}

    Help for data types. This is used by `get_help_for_data_types()`. Change this with `set_help_for_type()`.

  iter_config_instances_to_be_saved(**kw: Unpack[SaveKwargs]) → Iterator[Config[object]]
    Parameters

      - config_instances – The settings to consider

      - ignore – Skip these settings

    Iterate over all given `config_instances` and expand all `DictConfig` instances into the `Config` instances they consist of. Sort the resulting list if `config_instances` is not a `list` or a `tuple`. Yield all `Config` instances which are not (directly or indirectly) contained in `ignore` and where `Config.wants_to_be_exported()` returns true.

  name: str

    The name which is used in the config file to call this command. Use an empty string to define a default command which is used if an undefined command is encountered. If this is not set `get_name()` returns the name of this class in lower case letters and underscores replaced by hyphens.

  parse_and_set_value(instance: Config[Any], value: str) → None

    Parse the given value str and assign it to the given instance by calling `Config.parse_and_set_value()` with `ConfigFile.config_id` of `config_file`. Afterwards call `UiNotifier.show_info()`.

  parse_key_and_set_value(key: str, value: str) → None

    Find the corresponding `Config` instance for `key` and pass it to `parse_and_set_value()`.

    Raises

      ParseException – if key is invalid or `parse_and_set_value()` raises a `ValueError`

  run(cmd: Sequence[str], **kw: Unpack[ParseSplittedLineKwargs]) → None

    Call `set_multiple()` if the first argument contains `KEY_VAL_SEP` otherwise `set_with_spaces()`.

    Raises

      ParseException – if something is wrong (no arguments given, invalid syntax, invalid key, invalid value)

  save(writer: FormattedWriter, **kw: Unpack[SaveKwargs]) → None
    Parameters

      - writer – The file to write to

      - no_multi (bool) – If true: treat `MultiConfig` instances like normal `Config` instances and only write their default value. If false: Separate `MultiConfig` instances and print them once for every `MultiConfig.config_ids`.

      - comments (bool) – If false: don’t write help for data types

    Iterate over all `Config` instances with `iter_config_instances_to_be_saved()`, split them into normal `Config` and `MultiConfig` and write them with `save_config_instance()`. But before that set `last_name` to None (which is used by `write_config_help()`) and write help for data types based on `get_help_for_data_types()`.

  save_config_instance(writer: FormattedWriter, instance: Config[object], config_id: ConfigId | None, **kw: Unpack[SaveKwargs]) → None
    Parameters

      - writer – The file to write to

      - instance – The config value to be saved

      - config_id – Which value to be written in case of a `MultiConfig`, should be `None` for a normal `Config` instance

      - comments (bool) – If true: call `write_config_help()`

    Convert the `Config` instance into a value str with `format_value()`, wrap it in quotes if necessary with `config_file.quote()` and write it to `writer`.

  classmethod set_help_for_type(t: type[object], help_text: str) → None

    `get_help_for_data_types()` is used by `save()` and `get_help()`. Usually it uses the `help` attribute of the class. But if the class does not have a `help` attribute or if you want a different help text you can set the help with this method.

    Parameters

      - t – The type for which you want to specify a help

      - help_text – The help for `t`. It is cleaned up in `get_help_for_data_types()` with `HelpFormatterWrapper.format_text()` depending on `self.config_file.formatter_class`.

  set_multiple(cmd: Sequence[str]) → None

    Process one line of the format `set key=value [key2=value2 ...]`

    Raises

      MultipleParseExceptions – if something is wrong (invalid syntax, invalid key, invalid value)

  set_with_spaces(cmd: Sequence[str]) → None

    Process one line of the format `set key [=] value`

    Raises

      ParseException – if something is wrong (invalid syntax, invalid key, invalid value)

  write_config_help(writer: FormattedWriter, instance: Config[Any], *, group_dict_configs: bool = True) → None
    Parameters

      - writer – The output to write to

      - instance – The config value to be saved

    Write a comment which explains the meaning and usage of this setting based on `Config.format_allowed_values_or_type()` and `Config.help`.

    Use `last_name` to write the help only once for all `Config` instances belonging to the same `DictConfig` instance.

class confattr.configfile.SortedEnum(value)

  Bases: `Enum`

  By default it is assumed that the values are defined in ascending order `ONE='one'; TWO='two'; THREE='three'`. If you want to define them in descending order `THREE='three'; TWO='two'; ONE='one'` you can pass `descending = True` to the subclass. Passing `descending` requires Python 3.10.0a4 or newer. On older versions it causes a `TypeError: __prepare__() got an unexpected keyword argument 'descending'`. This was fixed in commit 6ec0adefad.

  descending: bool

class confattr.configfile.TextIOWriter(f: TextIO | None)

  Bases: `FormattedWriter`

  write_line_raw(ln: str) → None

class confattr.configfile.UiNotifier(notification_level: Union[Config[NotificationLevel], NotificationLevel] = NotificationLevel.ERROR)

  Bases: `object`

  Most likely you will want to load the config file before creating the UI (user interface). But if there are errors in the config file the user will want to know about them. This class takes the messages from `ConfigFile` and stores them until the UI is ready. When you call `set_ui_callback()` the stored messages will be forwarded and cleared.

  This object can also filter the messages. `ConfigFile` calls `show_info()` every time a setting is changed. If you load an entire config file this can be many messages and the user probably does not want to see them all. Therefore this object drops all messages of `NotificationLevel.INFO` by default. Pass `notification_level` to the constructor if you don’t want that.

  property notification_level: NotificationLevel

    Ignore messages that are less important than this level.

  set_ui_callback(callback: Callable[[NotificationLevel, str | BaseException], None]) → None

    Call `callback` for all messages which have been saved by `show()` and clear all saved messages afterwards. Save `callback` for `show()` to call.

  show(notification_level: NotificationLevel, msg: str | BaseException, *, ignore_filter: bool = False) → None

    If a callback for the user interface has been registered with `set_ui_callback()` call that callback. Otherwise save the message so that `set_ui_callback()` can forward the message when `set_ui_callback()` is called.

    Parameters

      - notification_level – The importance of the message

      - msg – The message to be displayed on the user interface

      - ignore_filter – If true: Show the message even if `notification_level` is smaller then the `UiNotifier.notification_level`.

  show_error(msg: str | BaseException, *, ignore_filter: bool = False) → None

    Call `show()` with `NotificationLevel.ERROR`.

  show_info(msg: str, *, ignore_filter: bool = False) → None

    Call `show()` with `NotificationLevel.INFO`.

class confattr.configfile.UnknownCommand(config_file: ConfigFile)

  Bases: `ConfigFileCommand`

  name: str = ''

    The name which is used in the config file to call this command. Use an empty string to define a default command which is used if an undefined command is encountered. If this is not set `get_name()` returns the name of this class in lower case letters and underscores replaced by hyphens.

  run(splitted_line: Sequence[str], **kw: Unpack[ParseSplittedLineKwargs]) → None

    Process one line which has been read from a config file

    Raises

      - ParseException – if there is an error in the line (e.g. invalid syntax)

      - MultipleParseExceptions – if there are several errors in the same line

confattr.configfile.readable_quote(value: str) → str

  This function has the same goal like `shlex.quote()` but tries to generate better readable output.

  Parameters

    value – A value which is intended to be used as a command line argument

  Returns

    A POSIX compliant quoted version of `value`